#!/usr/bin/env python3
from sysrepo_api import sysrepo_api3 as api
from pydbus import SystemBus

class SysrepoPortMirroring(api.SysrepoAPI):

    def __init__(self,
                 app_name="portmirroring",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/port-mirroring:port-mirroring",
                 indexes={"": "name"},
                 run_as_subscriber=False,
                 priority=0
                 ):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            run_as_subscriber=run_as_subscriber,
            indexes=indexes,
            priority=priority#,
            #service_list=service_list
        )
        self.bus = None
        self.pm = None

    def run_subscriber(self):
        self.__conn_pm_dbus()
        api.SysrepoAPI.run_subscriber(self)

    def add_subscribers(self):
        if self.module_name:
            self.subscriber.subscribe_module_change(module=self.module_name, xpath=None, callback=self.module_change_cb, priority=self.priority)

        self.logger.info("========== READING STARTUP CONFIG ===========")
        if self.namespace:
            for item in api.get_current_config_list(self.session, namespace=self.namespace):
                self.logger.debug(item)
            for item in api.get_current_config_list(self.session, namespace="/ucpe:config/interfaces:interface"):
                self.logger.debug(item)
    
    def __conn_pm_dbus(self):
        if self.pm != None:
            return

        try:
            self.bus = SystemBus()
            if self.bus is not None:
                self.pm = self.bus.get('net.telco.PortMirrorChange')
        except:
            self.logger.error('Connecting to \'net.telco.PortMirrorChange\' failed !')

    def __pm_check_vswitch(self, ifname, if_change_set, pm_changes):
        self.logger.info('Check vswitch secondary interfaces configuration for interface {}' .format(ifname))
        self.__conn_pm_dbus()
        
        oper = if_change_set.get('vswitch').get('operation', None)
        n_vsw = None
        o_vsw = None

        if oper == 'CREATED':
            n_vsw = if_change_set.get('vswitch').get('new_value', None)
        elif oper == 'MODIFIED':
            n_vsw = if_change_set.get('vswitch').get('new_value', None)
            o_vsw  = if_change_set.get('vswitch').get('old_value', None)
        elif oper == 'DELETED':
            o_vsw  = if_change_set.get('vswitch').get('old_value', None)
        
        config_indexed = self.get_port_mirrors()
        for name, config in sorted(config_indexed.items()):
            if 'src-port' in config:
                if o_vsw == config.get('src-port', '') or n_vsw == config.get('src-port', ''):
                    if self.pm is not None:
                        if (self.pm.set_src(name, config.get('src-port', '')) > 0):
                            self.logger.error("Failed to configure mirroring source")

    def __pm_check_vxlanif(self, ifname, if_change_set, pm_changes):
        self.logger.info('Check vxlan secondary interfaces configuration for interface {}' .format(ifname))
        self.__conn_pm_dbus()

        n_vxlanif = ifname

        if 'name' in if_change_set.keys():
            oper = if_change_set.get('name').get('operation', None)
            if oper == 'CREATED':
                n_vxlanif = ifname
            elif oper == 'DELETED':
                n_vxlanif  = None
        elif 'vni ' in if_change_set.keys() or 'local' in if_change_set.keys() \
                    or 'remote' in if_change_set.keys() or 'tunnel-type' in if_change_set.keys():
            n_vxlanif = ifname
        else:
            return
        
        config_indexed = self.get_port_mirrors()
        for name, config in sorted(config_indexed.items()):
            if 'src-port' in config:
                if ifname == config.get('src-port', ''):
                    if self.pm is not None:
                        if n_vxlanif == None:
                            if (self.pm.unset_src(name) > 0):
                                self.logger.error("Failed to configure mirroring source")
                        else:
                            if (self.pm.set_src(name, n_vxlanif) > 0):
                                self.logger.error("Failed to configure mirroring source")

    def __config_mirror(self, name, config, change_set):
        if self.get_network_mode() != 'vpp':
            self.logger.error('Skip port mirroring configuration - uCPE is not in VPP mode.')
            return

        if not change_set and config:
            # Startup config case - not in here
            return True

        self.__conn_pm_dbus()

        if change_set:
            self.logger.debug("APPLYING RUNNING FOR PORT MIRROR - {}".format(name))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))
            
            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        self.logger.debug("port mirror created - {}".format(name))
                        if self.pm is not None:
                            if (self.pm.create_mirror(name) > 0):
                                self.logger.error("Failed to create a new mirror")
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        self.logger.debug("port mirror deleted - return - {}".format(name))
                        if self.pm is not None:
                            if (self.pm.delete_mirror(name) > 0):
                                self.logger.error("Failed to delete mirror")
                        return
            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                n_status = False

                self.logger.debug('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('enable').get('new_value', False) and \
                            'true' in str(change_set.get('enable').get('new_value', False)).lower():
                        n_status = True
                    
                if self.pm is not None:
                    if (self.pm.set_enable(name, n_status) > 0):
                        self.logger.error("Failed to enable/disable mirror")
            
            if 'direction' in ch_set_keys:
                oper = change_set.get('direction').get('operation', None)
                n_dir = 'rxtx'

                self.logger.debug('Traffic direction is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_dir = change_set.get('direction').get('new_value', 'rxtx')
                    
                if self.pm is not None:
                    if (self.pm.set_direction(name, n_dir) > 0):
                        self.logger.error("Failed to configure traffic direction")

            if 'src-port' in ch_set_keys:
                oper = change_set.get('src-port').get('operation', None)
                n_src = None

                self.logger.debug('Mirroring destination is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_src = change_set.get('src-port').get('new_value', None)
                    
                if self.pm is not None:
                    if (self.pm.set_src(name, n_src) > 0):
                        self.logger.error("Failed to configure mirroring source")

            if 'dst-port' in ch_set_keys:
                oper = change_set.get('dst-port').get('operation', None)
                n_dst = None

                self.logger.debug('Mirroring destination is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    n_dst = change_set.get('dst-port').get('new_value', None)
                    
                if self.pm is not None:
                    if (self.pm.set_dst(name, n_dst) > 0):
                        self.logger.error("Failed to configure mirroring destination")

    def apply_config(self, change_set=None, apply_startup=False):
        config_indexed = self.get_port_mirrors()
        if apply_startup:
            # for name, config in sorted(config_indexed.items()):
            #     self.__config_mirror(name, config, change_set)
            return True

        pm_changes = {}
        interface_changes = {}
        for change in change_set:

            if change["xpath"].startswith(self.namespace):
                name = str(change['xpath'])
                name = name[name.find("port-mirroring[name='") + len("port-mirroring[name='"):]
                name = name[:name.find("]") - 1]
                if name not in pm_changes:
                    pm_changes[name] = {}

                pm_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change

            elif change["xpath"].startswith("/ucpe:config/interfaces:interface"):

                name = str(change['xpath'])
                name = name[name.find("interface[name='") + len("interface[name='"):]
                name = name[:name.find("]") - 1]
                if name not in interface_changes:
                    interface_changes[name] = {}

                interface_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change
            else:
                continue

        for name, xpath_dict in pm_changes.items():
            if name in config_indexed:
                self.__config_mirror(name, config_indexed[name], xpath_dict)
            else:
                self.__config_mirror(name, None, xpath_dict)
        
        for name, xpath_dict in interface_changes.items():
            if 'vswitch' in xpath_dict.keys():
                self.__pm_check_vswitch(name, xpath_dict, pm_changes)
            if 'vxlan' in name:
                self.__pm_check_vxlanif(name, xpath_dict, pm_changes)
        return True

    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")

    def set_port_mirror(self, name):
        self.set_item("{}[name='{}']".format(self.namespace, name), name)

    def get_port_mirror(self, name):
        return self.get_item("{}[name='{}']".format(self.namespace, name))

    def get_port_mirrors(self, startup=False):
        config_indexed = {}
        for mirror in self.get_startup_config(full_path=False) if startup else self.get_running_config(full_path=False):
            config_indexed[mirror['name']] = mirror
        return config_indexed

    def unset_port_mirror(self, name):
        self.delete_item("{}[name='{}']".format(self.namespace, name))

    def set_port_mirror_param(self, name, param, value):
        self.set_item("{}[name='{}']/{}".format(self.namespace, name, param), value)

    def get_port_mirror_param(self, name, param, default=None):
        try:
            for mirror in self.get_running_config(full_path=False):
                if mirror['name'] == name and param in mirror:
                    return self.get_item("{}[name='{}']/{}".format(self.namespace, name, param))
        except:
            return default

    def unset_port_mirror_param(self, name, param):
        self.delete_item("{}[name='{}']/{}".format(self.namespace, name, param))
